/* $Id: derivace.c,v 3.10 1998/05/23 22:33:10 kas Exp $ */

/* Rutiny pro vypocet derivace potencialove funkce */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995-1997  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <stdio.h>

#include "ff.h"
#include "linalg.h"
#include "pval.h"

/*
 * d(rk*(norm(a-b)-req)^2)/d(a1) =
 *    2*(a1-b1)*rk*(norm(a-b)-req)/norm(a-b) 
 */
void diff_bond(ffp, dx, dy, dz, x, y, z, bond)
force_field *ffp;
REAL *dx, *dy, *dz;
REAL *x, *y, *z;
int bond;
{
	int a1 = ffp->bondat1[bond];
	int a2 = ffp->bondat2[bond];

	REAL v12[3], norm, base;

	v12[0] = x[a1]-x[a2]; v12[1] = y[a1]-y[a2]; v12[2] = z[a1]-z[a2]; 
	norm = norm3(v12);
	base = 2 * ffp->rk[ffp->bondnum[bond]]
		* (norm - ffp->req[ffp->bondnum[bond]]) / norm;

	/* We hopefully have common subexpression optimizations ... */
	dx[a1] += v12[0]*base;
	dx[a2] -= v12[0]*base;
	dy[a1] += v12[1]*base;
	dy[a2] -= v12[1]*base;
	dz[a1] += v12[2]*base;
	dz[a2] -= v12[2]*base;
}

void diff_angle(ffp, dx, dy, dz, x, y, z, angle)
force_field *ffp;
REAL *dx, *dy, *dz;
REAL *x, *y, *z;
int angle;
{
	REAL v12[3];
	REAL v32[3];
	REAL common;

	REAL norm12, norm32, v12dotv32;

	int a1 = ffp->angleat1[angle];
	int a2 = ffp->angleat2[angle];
	int a3 = ffp->angleat3[angle];

	v12[0] = x[a1]-x[a2]; v12[1] = y[a1]-y[a2]; v12[2] = z[a1]-z[a2];
	v32[0] = x[a3]-x[a2]; v32[1] = y[a3]-y[a2]; v32[2] = z[a3]-z[a2];

	norm12 = norm3(v12);
	norm32 = norm3(v32);
	v12dotv32 = dot3(v12,v32);

	common = -2 * ffp->tk[ffp->anglenum[angle]] *
		(acos(v12dotv32/(norm12*norm32))
			- ffp->teq[ffp->anglenum[angle]])
		/ sqrt(norm12*norm12*norm32*norm32-v12dotv32*v12dotv32);

	dx[a1] += common*(x[a3]-x[a2]-v12dotv32*(x[a1]-x[a2])/(norm12*norm12));
	dy[a1] += common*(y[a3]-y[a2]-v12dotv32*(y[a1]-y[a2])/(norm12*norm12));
	dz[a1] += common*(z[a3]-z[a2]-v12dotv32*(z[a1]-z[a2])/(norm12*norm12));

	dx[a3] += common*(x[a1]-x[a2]-v12dotv32*(x[a3]-x[a2])/(norm32*norm32));
	dy[a3] += common*(y[a1]-y[a2]-v12dotv32*(y[a3]-y[a2])/(norm32*norm32));
	dz[a3] += common*(z[a1]-z[a2]-v12dotv32*(z[a3]-z[a2])/(norm32*norm32));

	dx[a2] -= common*(x[a3]-2*x[a2]+x[a1]
		- v12dotv32*(x[a1]-x[a2])/(norm12*norm12)
		- v12dotv32*(x[a3]-x[a2])/(norm32*norm32));
	dy[a2] -= common*(y[a3]-2*y[a2]+y[a1]
		- v12dotv32*(y[a1]-y[a2])/(norm12*norm12)
		- v12dotv32*(y[a3]-y[a2])/(norm32*norm32));
	dz[a2] -= common*(z[a3]-2*z[a2]+z[a1]
		- v12dotv32*(z[a1]-z[a2])/(norm12*norm12)
		- v12dotv32*(z[a3]-z[a2])/(norm32*norm32));
}

void diff_dih(ffp, dx, dy, dz, x, y, z, dih)
force_field *ffp;
REAL *dx, *dy, *dz;
REAL *x, *y, *z;
int dih;
{
	int a1, a2, a3, a4;
	REAL v21[3], v23[3], v34[3], v13[3], v24[3];
	REAL v123[3], v234[3], v123xv234[3];
	REAL cos_phi, sin_phi, n23, n123, n234, n123n234;
	REAL dephi, dphiv123[3], dphiv234[3];
	int n, dihnum;

	/* The following code is copied from compute_dih() in nontrans.c */
	a1 = ffp->dihat1[dih];
	a2 = ffp->dihat2[dih];
	a3 = ffp->dihat3[dih];
	a4 = ffp->dihat4[dih];
	dihnum = ffp->dihnum[dih];

        vectpts(v21,x,y,z,a1,a2);
        vectpts(v23,x,y,z,a3,a2);
        cross3(v123,v21,v23);
        n123 = norm3(v123);

        /* Normalovy vektor k rovine 2-3-4 */
        vectpts(v34,x,y,z,a4,a3);
        cross3(v234,v23,v34);
        n234 = norm3(v234);

	cross3(v123xv234,v123,v234);
	n23 = norm3(v23);
	if ((n123n234 = n123*n234)==0)
		return;
	cos_phi = -dot3(v123,v234)/n123n234;
	sin_phi = -dot3(v23,v123xv234)/n23/n123n234;

	/*
	 * Now in cos_phi we have a value of cos(<(1-2-3,2-3-4)).
	 * The following code makes diff cos(Pn*<(1-2-3,2-3-4)) from it.
	 */

	dephi = 0.0;
	do {
		REAL dcosnphi;
		if ((n=ffp->pn[dihnum]) < 0) {
			n = -n;
		}
		switch (n)
		{
		case 1: dcosnphi = 1;
			break;
		case 2:	dcosnphi = 4*cos_phi;
			break;
		case 3:	dcosnphi = 12*cos_phi*cos_phi - 3;
			break;
		case 4: dcosnphi = 32*cos_phi*cos_phi*cos_phi - 16*cos_phi;
			break;
		default:
			fprintf(stderr, "FATAL: in diff_dih(): Illegal periodicity (%d).\n",n);
			exit(1);
		}

		dephi += ffp->pk[dihnum]*ffp->phase[dihnum]*dcosnphi;
	} while (ffp->pn[dihnum++] < 0);

	dephi *= -sin_phi;

	vectpts(v13,x,y,z,a1,a3);
	vectpts(v24,x,y,z,a2,a4);

	cross3(dphiv123, v123, v23);
	scalmul3(dphiv123, 1/(n123*n123*n23));

	cross3(dphiv234, v234, v23);
	scalmul3(dphiv234, -1/(n234*n234*n23));

	dx[a1] += dephi * (v23[1]*dphiv123[2] - v23[2]*dphiv123[1]);
	dy[a1] += dephi * (v23[2]*dphiv123[0] - v23[0]*dphiv123[2]);
	dz[a1] += dephi * (v23[0]*dphiv123[1] - v23[1]*dphiv123[0]);

	dx[a2] += dephi * (v13[1]*dphiv123[2] - v13[2]*dphiv123[1]
			+ v34[2]*dphiv234[1] - v34[1]*dphiv234[2]);
	dy[a2] += dephi * (v13[2]*dphiv123[0] - v13[0]*dphiv123[2]
			+ v34[0]*dphiv234[2] - v34[2]*dphiv234[0]);
	dz[a2] += dephi * (v13[0]*dphiv123[1] - v13[1]*dphiv123[0]
			+ v34[1]*dphiv234[0] - v34[0]*dphiv234[1]);

	dx[a3] += dephi * (v21[2]*dphiv123[1] - v21[1]*dphiv123[2]
			+ v24[2]*dphiv234[1] - v24[1]*dphiv234[2]);
	dy[a3] += dephi * (v21[0]*dphiv123[2] - v21[2]*dphiv123[0]
			+ v24[0]*dphiv234[2] - v24[2]*dphiv234[0]);
	dz[a3] += dephi * (v21[1]*dphiv123[0] - v21[0]*dphiv123[1]
			+ v24[1]*dphiv234[0] - v24[0]*dphiv234[1]);

	dx[a4] += dephi * (v23[2]*dphiv234[1] - v23[1]*dphiv234[2]);
	dy[a4] += dephi * (v23[0]*dphiv234[2] - v23[2]*dphiv234[0]);
	dz[a4] += dephi * (v23[1]*dphiv234[0] - v23[0]*dphiv234[1]);

	if (!ffp->ignored[dih])
		 diff_14(ffp, dx,dy,dz, x,y,z, a1,a4);
}

void diff_14(ffp, dx, dy, dz, x, y, z, a1, a2)
force_field *ffp;
REAL *dx, *dy, *dz;
REAL *x, *y, *z;
int a1, a2;
{
	int idx;
	REAL dist, d2, common;

	idx = ffp->nbnum[ffp->natypes*ffp->atypes[a1]+ffp->atypes[a2]];
	if (idx < 0) {
		int tsmall, tbig, foo;
		tsmall=ffp->atypes[a1];
		tbig=ffp->atypes[a2];
		if (tsmall > tbig) {
			foo = tsmall; tsmall = tbig; tbig = foo;
		}
		idx = (tbig*(tbig+1))/2 + tsmall;
	}

	d2 = compute_distance_sq(x,y,z, a1, a2);
	dist = sqrt(d2);

	common = (-12*ffp->r12[idx]/(d2*d2*d2*d2*d2*d2*d2)
		+6*ffp->r6[idx]/(d2*d2*d2*d2)) / ffp->vdw_14;

	/* Electrostatic interaction: */
	common -= ((ffp->diel_const+2*ffp->diel_lin*dist)
		* ffp->charge[a1]*ffp->charge[a2]
		/ ((ffp->diel_lin*dist+ffp->diel_const)*
		   (ffp->diel_lin*dist+ffp->diel_const)*dist*d2))
		/ ffp->eel_14;
	dx[a1] += (x[a1]-x[a2])*common;
	dx[a2] -= (x[a1]-x[a2])*common;
	dy[a1] += (y[a1]-y[a2])*common;
	dy[a2] -= (y[a1]-y[a2])*common;
	dz[a1] += (z[a1]-z[a2])*common;
	dz[a2] -= (z[a1]-z[a2])*common;
}


void diff_nb(ffp, dx, dy, dz, x, y, z, a1, a2)
force_field *ffp;
REAL *dx, *dy, *dz;
REAL *x, *y, *z;
int a1, a2;
{
	int idx;
	REAL dist, d2, common;

	idx = ffp->nbnum[ffp->natypes*ffp->atypes[a1]+ffp->atypes[a2]];

	d2 = compute_distance_sq(x,y,z, a1, a2);
	dist = sqrt(d2);

	/*
	 * Distinguish between a VdW (12-6) interaction and
	 * the hydrogen bond (12-10 interaction).
	 */
	if (idx >= 0)
		common = -12*ffp->r12[idx]/(d2*d2*d2*d2*d2*d2*d2)
			+6*ffp->r6[idx]/(d2*d2*d2*d2);
	else
		common = -12*ffp->h12[-idx-2]/(d2*d2*d2*d2*d2*d2*d2)
			+10*ffp->h10[-idx-2]/(d2*d2*d2*d2*d2*d2);
	/* Electrostatic interaction: */
	common -= (ffp->diel_const+2*ffp->diel_lin*dist)
		* ffp->charge[a1]*ffp->charge[a2]
		/ ((ffp->diel_lin*dist+ffp->diel_const)*
		   (ffp->diel_lin*dist+ffp->diel_const)*dist*d2);

	dx[a1] += (x[a1]-x[a2])*common;
	dx[a2] -= (x[a1]-x[a2])*common;
	dy[a1] += (y[a1]-y[a2])*common;
	dy[a2] -= (y[a1]-y[a2])*common;
	dz[a1] += (z[a1]-z[a2])*common;
	dz[a2] -= (z[a1]-z[a2])*common;
}

/*
 * The following routine is an equivalent of compute_potential()
 * for computing gradient.
 */
void compute_gradient(pvp)
potential_value *pvp;
{
	int i,j;

	for (i=0; i<3*pvp->ffp->natoms; i++)
		pvp->dx[i] = 0.0;
#if !defined(ANGLES_ONLY) && !defined(TORSIONS_ONLY) && !defined(NBOND_ONLY)
	for (i=0; i<pvp->ffp->nbonds; i++)
		diff_bond(pvp->ffp,
			pvp->dx, pvp->dy, pvp->dz, pvp->x, pvp->y, pvp->z, i);
#endif
#if !defined(BONDS_ONLY) && !defined(TORSIONS_ONLY) && !defined(NBOND_ONLY)
	for (i=0; i<pvp->ffp->nangles; i++) {
		diff_angle(pvp->ffp,
			pvp->dx, pvp->dy, pvp->dz, pvp->x, pvp->y, pvp->z, i);
	}
#endif
#if !defined(BONDS_ONLY) && !defined(ANGLES_ONLY) && !defined(NBOND_ONLY)
	for (i=0; i<pvp->ffp->ndihs; i++)
		diff_dih(pvp->ffp,
			pvp->dx, pvp->dy, pvp->dz, pvp->x, pvp->y, pvp->z, i);
#endif
#if !defined(BONDS_ONLY) && !defined(ANGLES_ONLY) && !defined(TORSIONS_ONLY)
	if (pvp->cutoff_lifetime < 0)
		recompute_cutoff(pvp);
        for (i=0; i<pvp->ffp->natoms; i++)
                for (j=0; j<pvp->nbonds[i][0]; j++)
                {
                        if (pvp->nbonds[i][1+j] <= i)
                                continue;
                                diff_nb(pvp->ffp, pvp->dx, pvp->dy, pvp->dz,
					pvp->x, pvp->y, pvp->z,
                                        i, pvp->nbonds[i][1+j]);
                }
#endif
}
