/* $Id: pval.c,v 3.16 1998/05/23 22:33:13 kas Exp $ */

/* Rutiny pro vypocet funkcni hodnoty potencialove funkce */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995-1997  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ff.h"
#include "linalg.h"
#include "pval.h"

/* The following macros are for debugging of single type of interaction */
#undef BONDS_ONLY
#undef ANGLES_ONLY
#undef TORSIONS_ONLY
#undef NBOND_ONLY

#ifdef PVAL_VERBOSE
int pval_verbose = PVAL_SUMVERBOSE;
#endif

#define max(x,y)	((x)<(y) ? (y) : (x))

potential_value *alloc_pv(f)
force_field *f;
{
	potential_value *pvp;
	int i,j,s, *tmp, *nb, mx;
	int maxangles = 0, maxbonds = 0, maxdihs = 0;

	pvp = malloc(sizeof(potential_value));

	pvp->ffp = f;

	pvp->nbval = malloc(sizeof(REAL)*f->natoms*f->natoms);
	/*
	 * Vynulujeme nbval, protoze v compute_cutoff() se pocita uz s platnymi
	 * hodnotami.
	 */
	bzero(pvp->nbval, sizeof(REAL)*f->natoms*f->natoms);
	pvp->bondval = malloc(sizeof(REAL)*f->nbonds);
	pvp->angleval = malloc(sizeof(REAL)*f->nangles);
	pvp->dihval = malloc(sizeof(REAL)*f->ndihs);

	/*
	 * Zmna: Vechny souadnice se nyn alokuj jako jedno
	 * pole, aby se to pak ppadn mohlo pedhodit obecn
	 * minimalizan procedue. Pro kompatibilitu zachovvm
	 * ti pointery x, y a z, ale malloc()uje se jen x.
	 */
	pvp->x = malloc(3*sizeof(REAL)*f->natoms);
	pvp->y = pvp->x+f->natoms;
	pvp->z = pvp->y+f->natoms;

	pvp->dx = malloc(3*sizeof(REAL)*f->natoms);
	pvp->dy = pvp->dx+f->natoms;
	pvp->dz = pvp->dy+f->natoms;

	pvp->nbonds = malloc(sizeof(int *)*f->natoms);
	pvp->bonds = malloc(sizeof(int *)*f->natoms);
	pvp->angles = malloc(sizeof(int *)*f->natoms);
	pvp->dihs = malloc(sizeof(int *)*f->natoms);

	/* Done allocating basic arrays, now we have to compute
	 * bonds, angles and dihs arrays for each atom.
	 */
	/*
	 * Nasledujici kod se castecne vyskytuje i v recompute_cutoff()
	 */
	mx = max(f->natoms, f->nbonds);
	mx = max(f->nangles, mx);
	mx = max(f->ndihs, mx);

	tmp = malloc(sizeof(int)*mx);
	nb = malloc(sizeof(int)*(f->natoms));

	for (i=0; i<f->natoms; i++)
	{
		for (j=0; j<f->natoms; j++)
			nb[j] = 1;
		nb[i] = 0;

		for (j=s=0; j<f->nbonds; j++)
			if (f->bondat1[j] == i || f->bondat2[j] == i)
			{
				tmp[s++] = j;
				nb[f->bondat1[j]] = 0;
				nb[f->bondat2[j]] = 0;
			}
		pvp->bonds[i] = malloc(sizeof(int)*(s+1));
		pvp->bonds[i][0] = s;
		if (s)
			memcpy(pvp->bonds[i]+1,tmp,s*sizeof(int));

		if (s > maxbonds)
			maxbonds = s;

		for (j=s=0; j<f->nangles; j++)
			if (f->angleat1[j] == i || f->angleat2[j] == i
				|| f->angleat3[j] == i)
			{
				tmp[s++] = j;
				nb[f->angleat1[j]] = 0;
				nb[f->angleat2[j]] = 0;
				nb[f->angleat3[j]] = 0;
			}
		pvp->angles[i] = malloc(sizeof(int)*(s+1));
		pvp->angles[i][0] = s;
		if (s)
			memcpy(pvp->angles[i]+1,tmp,s*sizeof(int));

		if (s > maxangles)
			maxangles = s;

		for (j=s=0; j<f->ndihs; j++)
			if (f->dihat1[j] == i || f->dihat2[j] == i ||
				f->dihat3[j] == i || f->dihat4[j] == i)
			{
				tmp[s++] = j;
				nb[f->dihat1[j]] = 0;
				nb[f->dihat2[j]] = 0;
				nb[f->dihat3[j]] = 0;
				nb[f->dihat4[j]] = 0;
			}
		pvp->dihs[i] = malloc(sizeof(int)*(s+1));
		pvp->dihs[i][0] = s;
		if (s)
			memcpy(pvp->dihs[i]+1,tmp,s*sizeof(int));

		if (s > maxdihs)
			maxdihs = s;

		/* Now find all nonbonded interactions for atom i */
		for(j=0, s=0; j<f->natoms; j++)
			if (nb[j])
				s++;
		pvp->nbonds[i] = malloc(sizeof(int)*(s+1));
		pvp->nbonds[i][0] = s;
		for(j=0, s=0; j<f->natoms; j++)
			if (nb[j])
				pvp->nbonds[i][++s] = j;
	}
	free(tmp);
	free(nb);
	pvp->recomputed_atom = -1;
	pvp->cutoff_lifetime = -1; /* Cutoff needs to be computed */

#ifdef DEBUG
	printf("maxbonds = %d, maxangles = %d, maxdihs = %d\n",
		maxbonds, maxangles, maxdihs);
#endif
	pvp->onbval = malloc(sizeof(REAL)*f->natoms);
	pvp->obondval = malloc(sizeof(REAL)*maxbonds);
	pvp->oangleval = malloc(sizeof(REAL)*maxangles);
	pvp->odihval = malloc(sizeof(REAL)*maxdihs);

	return pvp;
}

void free_pv(pvp)
potential_value *pvp;
{
	int i;

	free(pvp->bondval);
	free(pvp->angleval);
	free(pvp->dihval);

	/*
	 * x, y a z jsou nyn alokovny jako jeden celek, take sta
	 * jen tento dek:
	 */
	free(pvp->x);
	/* free(pvp->y);
	   free(pvp->z); */
	free(pvp->dx);
	/* free(pvp->dy);
	   free(pvp->dz); */

	free(pvp->obondval);
	free(pvp->oangleval);
	free(pvp->odihval);

	for (i=0; i<pvp->ffp->natoms; i++)
	{
		free(pvp->bonds[i]);
		free(pvp->angles[i]);
		free(pvp->dihs[i]);
	}

	free(pvp->bonds);
	free(pvp->angles);
	free(pvp->dihs);

	free_ff(pvp->ffp);
}

/*
 * Funkce recompute_cutoff ma za ukol odecist od hodnoty siloveho
 * pole hodnoty nevazebnych interakci, spocitat, ktere nevazebne
 * interakce se nadale maji brat v uvahu na zaklade cutoffu,
 * a spocitat hodnotu nevazebnych interakci.
 * Nasledujici kod se castecne vyskytuje i v alloc_pv().
 */
void recompute_cutoff(pvp)
potential_value *pvp;
{
	int i, j, s, res_i, res_j, idx;
	int *nb;
	REAL *res_distance, r;

	force_field *f = pvp->ffp;
	/*
	 * Nejprve odecteme hodnoty starych nevazebnych interakci:
	 * Tohle by mela byt nula (anebo by na tom nemelo zalezet)
	 * u prvniho vypoctu potencialove funkce.
	 * FIXME: Tohle jeste promyslet.
	 */
	for (i=0; i<pvp->ffp->natoms; i++) {
/* Vyrazeno 9.6.97 Petr Mejzlik
		for (j=0; j<pvp->nbonds[i][0]; j++) {
			if (pvp->nbonds[i][1+j] <= i)
				continue;
			pvp->value -=
				pvp->nbval[pvp->ffp->natoms*i+pvp->nbonds[i][1+j]];
		}
*/
		free (pvp->nbonds[i]);
	}

	/*
	 * Ted spocitame minimalni vzdalenosti jednotlivych residui.
	 * POZOR! Nasledujici radek muze volat malloc(0) pro jedno
	 * residuum.
	 */
	if (f->nresidues > 1)
		res_distance = malloc(sizeof(REAL)*f->nresidues*(f->nresidues-1)/2);
	else /* just to keep the compiler happy */
		res_distance = 0;

	for (i=0; i<f->nresidues*(f->nresidues-1)/2; i++)
		res_distance[i] = 2.0*f->cutoff_sq;

	for (i=0,res_i=0,idx=0; i<f->natoms-1; i++)
	{
		if (f->residue[res_i] <= i)
			res_i++; 
		for (j=i,res_j=res_i; j<f->natoms; j++)
		{
			if (f->residue[res_j] <= j)
				res_j++;
			if (res_i == res_j)
				continue;
			idx = res_j*(res_j-1)/2+res_i;
			if (res_distance[idx] < f->cutoff_sq)
				continue;
			if ((r=compute_distance_sq(pvp->x,pvp->y,pvp->z, i,j))
				< f->cutoff_sq)
				res_distance[idx]=r;
		}
	}

#ifdef DEBUG_RESIDUE
	for (i=0; i<f->nresidues-1; i++) {
		int j;
		for (j=i+1;j<f->nresidues; j++)
			printf("dist(%d,%d) = %f\n", i,j, res_distance[j*(j-1)/2+i]);
	}
#endif
/*
 * Ted spocitame pro kazdy atom mnozinu jeho nevazebnych interakci.
 */
	nb = malloc(sizeof(int)*(f->natoms));

	for (i=res_i=0; i<f->natoms; i++)
	{
		if (f->residue[res_i] <= i)
			res_i++;

		/*
		 * Nejprve zahrneme vsechny atomy z vlastniho residua
		 * a z jinych residui, ktere jsou blize nez cutoff.
		 */
		for (j=0, res_j=0; j<f->natoms; j++)
		{
			if (f->residue[res_j] <= j)
				res_j++;
			if (res_i == res_j) {
				nb[j] = 1;
				continue;
			}
			if (res_i < res_j && res_distance[res_j*(res_j-1)/2
				+res_i] < f->cutoff_sq) {
				nb[j] = 1;
				continue;
			}
			if (res_i > res_j && res_distance[res_i*(res_i-1)/2
				+res_j] < f->cutoff_sq) {
				nb[j] = 1;
				continue;
			}
			nb[j] = 0;
		}
		nb[i] = 0;

		/* Nyni vyskrtame vsechny vazby pro atom i */
		for (j=0; j<f->nbonds; j++)
			if (f->bondat1[j] == i || f->bondat2[j] == i)
			{
				nb[f->bondat1[j]] = 0;
				nb[f->bondat2[j]] = 0;
			}
		/* Nyni vyskrtame vsechny uhly pro atom i */
		for (j=0; j<f->nangles; j++)
			if (f->angleat1[j] == i || f->angleat2[j] == i
				|| f->angleat3[j] == i)
			{
				nb[f->angleat1[j]] = 0;
				nb[f->angleat2[j]] = 0;
				nb[f->angleat3[j]] = 0;
			}
		/* Nyni vyskrtame vsechny torze pro atom i */
		for (j=0; j<f->ndihs; j++)
			if (f->dihat1[j] == i || f->dihat2[j] == i ||
				f->dihat3[j] == i || f->dihat4[j] == i)
			{
				nb[f->dihat1[j]] = 0;
				nb[f->dihat2[j]] = 0;
				nb[f->dihat3[j]] = 0;
				nb[f->dihat4[j]] = 0;
			}
		/* A ted na zaklade pole nb vypocteme polozku pvp->nbonds */
		for (j=s=0; j<f->natoms; j++)
			if (nb[j]) s++;
		pvp->nbonds[i] = malloc(sizeof(int)*(s+1));
		pvp->nbonds[i][0] = s;
		for(j=0, s=0; j<f->natoms; j++)
			if (nb[j])
				pvp->nbonds[i][++s] = j;
	}
	free(nb);
	pvp->recomputed_atom = -1;
	pvp->cutoff_lifetime = CUTOFF_LIFETIME;
}

int read_coordinates(pvp, filename)
potential_value *pvp;
char *filename;
{
	FILE *f;

	int i;

	if ((f = fopen(filename, "r")) == NULL)
		return 0;
	
	for (i=0; i<pvp->ffp->natoms; i++)
		if (fscanf(f, "%lf %lf %lf \n",
			pvp->x+i, pvp->y+i, pvp->z+i) != 3)
		{
			fclose(f);
			return 0;
		}

	fclose(f);

	/* After the new coordinates is read recompute the cutoff. */
	pvp->cutoff_lifetime = -1;
	return 1;
}

REAL compute_potential(pvp)
potential_value *pvp;
{
	int i, j;
#ifdef PVAL_VERBOSE
	REAL r = 0.0;
#endif
	pvp->value = 0.0;

#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_DETAILED)
		puts("####  At1- At2 Type   Rk       Req   Length           Energy");
#endif
#if !defined(ANGLES_ONLY) && !defined(TORSIONS_ONLY) && !defined(NBOND_ONLY)
	for (i=0; i<pvp->ffp->nbonds; i++)
		pvp->value += pvp->bondval[i] = compute_bond(pvp->ffp,
			pvp->x, pvp->y, pvp->z, i);
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_INTSUM) {
		printf ("#### Bonds total:                            %15.3f\n",
			pvp->value);
		r = pvp->value;
	}
	if (pval_verbose >= PVAL_DETAILED)
		puts("####  At1- At2- AT3 Type   Tk      Teq    Angle       Energy");
#endif
#endif
#if !defined(BONDS_ONLY) && !defined(TORSIONS_ONLY) && !defined(NBOND_ONLY)
	for (i=0; i<pvp->ffp->nangles; i++) {
		pvp->value += pvp->angleval[i] = compute_angle(pvp->ffp,
			pvp->x, pvp->y, pvp->z, i);
	}
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_INTSUM) {
		printf ("#### Angles total:                           %15.3f\n",
			pvp->value-r);
		r = pvp->value;
	}
	if (pval_verbose >= PVAL_DETAILED) {
		puts("####  At1- At2- At3- At4 Type Phase    Pk    T-Angle  Energy");
		puts("####  At1- At4       R12       R6     Dist     VdW       EEL");
	}
#endif
#endif
#if !defined(BONDS_ONLY) && !defined(ANGLES_ONLY) && !defined(NBOND_ONLY)
	for (i=0; i<pvp->ffp->ndihs; i++)
		pvp->value += pvp->dihval[i] = compute_dih(pvp->ffp,
			pvp->x, pvp->y, pvp->z, i);
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_INTSUM) {
		printf ("#### Dihedrals total (incl. 1-4 interactions): %13.3f\n",
			pvp->value-r);
		r = pvp->value;
	}
	if (pval_verbose >= PVAL_DETAILED)
		puts("####  At1- At2      R12/H12   R6/H10   Dist    VdW/H     EEL");
#endif
#endif
#if !defined(BONDS_ONLY) && !defined(ANGLES_ONLY) && !defined(TORSIONS_ONLY)
	if (pvp->cutoff_lifetime < 0)
		recompute_cutoff(pvp);
	if (pvp->ffp->cutoff_sq > 0)
		pvp->cutoff_lifetime--;
	for (i=0; i<pvp->ffp->natoms; i++)
		for (j=0; j<pvp->nbonds[i][0]; j++)
		{
			if (pvp->nbonds[i][1+j] <= i)
				continue;
			pvp->value +=
			pvp->nbval[pvp->ffp->natoms*i+pvp->nbonds[i][1+j]] =
				compute_nb(pvp->ffp, pvp->x, pvp->y, pvp->z,
					i, pvp->nbonds[i][1+j]);
		}
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_INTSUM)
		printf ("#### Nonbonded/HBond interactions total:     %15.3f\n",
			pvp->value-r);
	if (pval_verbose >= PVAL_SUMVERBOSE)
		printf("#### Total value of potential energy:        ");
	if (pval_verbose >= PVAL_SUM)
		printf("%15.3f\n", pvp->value);

#endif
#endif
	pvp->recomputed_atom = -1;

	return pvp->value;
}

REAL recompute_potential(pvp, atom, nx, ny, nz)
potential_value *pvp;
int atom;
REAL nx, ny, nz;
{
	int i;

	pvp->ox = pvp->x[atom];
	pvp->oy = pvp->y[atom];
	pvp->oz = pvp->z[atom];

	pvp->x[atom] = nx;
	pvp->y[atom] = ny;
	pvp->z[atom] = nz;

	pvp->recomputed_atom = atom;
	pvp->ovalue = pvp->value;

	/* Recompute all bonds, angles, dihedrals for atom */
#if !defined(ANGLES_ONLY) && !defined(TORSIONS_ONLY) && !defined(NBOND_ONLY)
	for (i=pvp->bonds[atom][0]; i; i--)
	{
		int bond;

		bond = pvp->bonds[atom][i];
		pvp->obondval[i-1] = pvp->bondval[bond];
		pvp->value -= pvp->bondval[bond];

		pvp->bondval[bond] = compute_bond(pvp->ffp,
			pvp->x, pvp->y, pvp->z, bond);
		pvp->value += pvp->bondval[bond];
	}
#endif
#if !defined(BONDS_ONLY) && !defined(TORSIONS_ONLY) && !defined(NBOND_ONLY)
	for (i=pvp->angles[atom][0]; i; i--)
	{
		int angle;

		angle = pvp->angles[atom][i];
		pvp->oangleval[i-1] = pvp->angleval[angle];
		pvp->value -= pvp->angleval[angle];
		pvp->angleval[angle] = compute_angle(pvp->ffp,
			pvp->x, pvp->y, pvp->z, angle);
		pvp->value += pvp->angleval[angle];
	}
#endif
#if !defined(BONDS_ONLY) && !defined(ANGLES_ONLY) && !defined(NBOND_ONLY)
	for (i=pvp->dihs[atom][0]; i; i--)
	{
		int dih;

		dih = pvp->dihs[atom][i];
		pvp->odihval[i-1] = pvp->dihval[dih];
		pvp->value -= pvp->dihval[dih];

		pvp->dihval[dih] = compute_dih(pvp->ffp,
			pvp->x, pvp->y, pvp->z, dih);
		pvp->value += pvp->dihval[dih];
	}
#endif
#if !defined(BONDS_ONLY) && !defined(ANGLES_ONLY) && !defined(TORSIONS_ONLY)
	for (i=pvp->nbonds[atom][0]; i; i--)
	{
		int a1,a2, idx;

		if ((a2 = pvp->nbonds[atom][i]) <= atom) {
			a1 = a2;
			a2 = atom;
		} else
			a1 = atom;
		
		idx = a1*pvp->ffp->natoms + a2;

		pvp->onbval[i-1] = pvp->nbval[idx];
#ifdef PRINT_NB_DEBUG
		printf("onbval=%f\n", pvp->nbval[idx]);
#endif
		pvp->value -= pvp->nbval[idx];

		pvp->nbval[idx] = compute_nb(pvp->ffp, pvp->x, pvp->y,
			pvp->z, a1, a2);
		pvp->value += pvp->nbval[idx];
#ifdef PRINT_NB_DEBUG
		printf("nbval=%f\n", pvp->nbval[idx]);
		printf("value = %f\n", pvp->value);
#endif
	}
#endif
	if (pvp->ffp->cutoff_sq > 0)
		pvp->cutoff_lifetime--;
	return pvp->value;
}

REAL undo_new_potential(pvp)
potential_value *pvp;
{
	int i;

	/* This is to discover calling of undo_new_potential() w/o previous
	 * call of recompute_potential().
	 */
	if (pvp->recomputed_atom == -1)
	{
		fprintf(stderr, "Oops! Trying to undo something that did not happen!\n");
		exit(1);
	}

	pvp->value = pvp->ovalue;

	pvp->x[pvp->recomputed_atom] = pvp->ox;
	pvp->y[pvp->recomputed_atom] = pvp->oy;
	pvp->z[pvp->recomputed_atom] = pvp->oz;

	for (i=pvp->bonds[pvp->recomputed_atom][0]; i; i--)
		pvp->bondval[pvp->bonds[pvp->recomputed_atom][i]] =
			pvp->obondval[i-1];

	for (i=pvp->angles[pvp->recomputed_atom][0]; i; i--)
		pvp->angleval[pvp->angles[pvp->recomputed_atom][i]] =
			pvp->oangleval[i-1];

	for (i=pvp->dihs[pvp->recomputed_atom][0]; i; i--)
		pvp->dihval[pvp->dihs[pvp->recomputed_atom][i]] =
			pvp->odihval[i-1];

	for (i=pvp->nbonds[pvp->recomputed_atom][0]; i; i--)
		pvp->nbval [pvp->nbonds[pvp->recomputed_atom][i]
			* pvp->ffp->natoms + pvp->recomputed_atom]
		= pvp->nbval[pvp->recomputed_atom * pvp->ffp->natoms
			+ pvp->nbonds[pvp->recomputed_atom][i]]
		= pvp->onbval[i-1];

	pvp->recomputed_atom = -1;

	return pvp->value;
}

